@@ -39,20 +39,20 @@ SKKCore::SKKCore      (KeyBind *keybind, SKKDictionaries *dict,
       m_dict(dict),
       m_skk_mode(SKK_MODE_HIRAGANA),
       m_input_mode(INPUT_MODE_DIRECT),
-      m_learning(0),
-      m_end_flag(false),
-      m_ltable(ltable),
-      m_show_ltable(false),
       m_key2kana(key2kana),
+      m_learning(0),
+      m_commit_flag(false),
+      m_end_flag(false),
       m_preedit_pos(0),
       m_commit_pos(0),
-      m_cindex(m_candlist.end()),
-      m_commit_flag(false)
+      m_show_ltable(false),
+      m_ltable(ltable),
+      m_cindex(m_candlist.end())
 {
     m_ltable->clear();
     clear_preedit();
     clear_commit();
-    clear_pending();
+    clear_pending(false);
 }
 
 SKKCore::~SKKCore     (void)
@@ -192,6 +192,8 @@ SKKCore::commit_converting (int index)
             m_dict->strip_annot(str);
         commit_string(str);
         commit_string(m_okuristr);
+        if (m_okurihead != 0)
+            m_preeditstr += m_okurihead;
         m_dict->write(m_preeditstr, *m_cindex);
         clear_candidate();
         clear_preedit();
@@ -205,6 +207,8 @@ SKKCore::commit_converting (int index)
             m_dict->strip_annot(cand);
         commit_string(cand);
         commit_string(m_okuristr);
+        if (m_okurihead != 0)
+            m_preeditstr += m_okurihead;
         m_dict->write(m_preeditstr, str);
         clear_candidate();
         clear_preedit();
@@ -236,6 +240,8 @@ SKKCore::caret_pos (void)
             base_pos += m_okuristr.length() + 1;
         return base_pos + m_preeditstr.length() + 2 + m_learning->caret_pos();
     }
+
+    return base_pos;
 }
 
 void
@@ -333,9 +339,9 @@ SKKCore::get_input_mode (void)
 
 
 void
-SKKCore::clear_pending (void)
+SKKCore::clear_pending (bool flag)
 {
-    if (m_pendingstr == utf8_mbstowcs("n")) {
+    if (flag && m_pendingstr == utf8_mbstowcs("n")) {
         commit_or_preedit(utf8_mbstowcs("\xE3\x82\x93"));
     }
     m_pendingstr.clear();
@@ -385,6 +391,7 @@ SKKCore::action_kakutei (void)
               m_skk_mode == SKK_MODE_WIDE_ASCII) &&
             m_pendingstr.empty() && m_preeditstr.empty()) {
             m_end_flag = true;
+            return false;
         } else {
             clear_pending();
         }
@@ -413,6 +420,8 @@ SKKCore::action_kakutei (void)
             commit_converting(0);
         set_input_mode(INPUT_MODE_DIRECT);
         break;
+    default:
+        break;
     }
     if(m_skk_mode == SKK_MODE_ASCII || m_skk_mode == SKK_MODE_WIDE_ASCII) {
         set_skk_mode(SKK_MODE_HIRAGANA);
@@ -427,7 +436,7 @@ SKKCore::action_cancel (void)
     switch (m_input_mode) {
     case INPUT_MODE_DIRECT:
         if (!m_pendingstr.empty()) {
-            clear_pending();
+            clear_pending(false);
         } else {
             clear_commit();
             m_end_flag = true;
@@ -437,7 +446,7 @@ SKKCore::action_cancel (void)
     case INPUT_MODE_PREEDIT:
     case INPUT_MODE_OKURI:
         clear_preedit();
-        clear_pending();
+        clear_pending(false);
         set_input_mode(INPUT_MODE_DIRECT);
         if (m_skk_mode == SKK_MODE_ASCII) {
             set_skk_mode(SKK_MODE_HIRAGANA);
@@ -453,6 +462,8 @@ SKKCore::action_cancel (void)
         set_input_mode(INPUT_MODE_PREEDIT);
         clear_candidate();
         break;
+    default:
+        break;
     }
     return retval;
 }
@@ -484,6 +495,8 @@ SKKCore::action_convert (void)
                 m_cindex = m_candlist.begin();
         }
         return true;
+    default:
+        break;
     }
 
     return false;
@@ -537,11 +550,36 @@ SKKCore::action_katakana (bool half)
             set_skk_mode(SKK_MODE_KATAKANA);
         }
         return true;
+    default:
+        break;
     }
 
     return false;
 }
 
+bool
+SKKCore::action_toggle_case (void)
+{
+    if (m_input_mode == INPUT_MODE_PREEDIT && m_skk_mode == SKK_MODE_ASCII) {
+        for (WideString::iterator i = m_preeditstr.begin();
+             i != m_preeditstr.end(); i++) {
+            int code = *i;
+            if (islower(code)) {
+                *i = toupper(code);
+            } else if (isupper(code)) {
+                *i = tolower(code);
+            }
+        }
+        commit_string(m_preeditstr);
+        clear_preedit();
+        clear_pending();
+        set_input_mode(INPUT_MODE_DIRECT);
+        set_skk_mode(SKK_MODE_HIRAGANA);
+        return true;
+    }
+    return false;
+}
+
 bool
 SKKCore::action_start_preedit (void)
 {
@@ -600,6 +638,8 @@ SKKCore::action_ascii (bool wide)
         else
             commit_converting(0);
         set_input_mode(INPUT_MODE_DIRECT);
+    default:
+        break;
     }
     clear_pending();
     if (wide) {
@@ -657,6 +697,8 @@ SKKCore::action_backspace (void)
                 m_commitstr.erase(m_commit_pos-1, 1);
                 m_commit_pos--;
             }
+        default:
+            break;
         }
     } else {
         if (m_input_mode == INPUT_MODE_OKURI &&
@@ -694,6 +736,8 @@ SKKCore::action_delete (void)
             } else if (m_commit_pos < m_commitstr.length()) {
                 m_commitstr.erase(m_commit_pos, 1);
             }
+        default:
+            break;
         }
     } else {
         clear_pending();
@@ -726,6 +770,8 @@ SKKCore::action_forward (void)
             return false;
         }
         return true;
+    default:
+        break;
     }
     return false;
 }
@@ -754,10 +800,70 @@ SKKCore::action_backward (void)
             return false;
         }
         return true;
+    default:
+        break;
     }
     return false;
 }
 
+bool
+SKKCore::action_home (void)
+{
+    switch (m_input_mode) {
+    case INPUT_MODE_CONVERTING:
+        return false;
+    case INPUT_MODE_DIRECT:
+        clear_pending();
+        if (m_commit_pos > 0) {
+            m_commit_pos = 0;
+            return true;
+        } else {
+            return false;
+        }
+    case INPUT_MODE_PREEDIT:
+        clear_pending();
+        if (m_preedit_pos > 0) {
+            m_preedit_pos = 0;
+        } else if (m_commit_pos > 0) {
+            m_commit_pos = 0;
+        } else {
+            return false;
+        }
+        return true;
+    default:
+        return false;
+    }
+}
+
+bool
+SKKCore::action_end (void)
+{
+    switch (m_input_mode) {
+    case INPUT_MODE_CONVERTING:
+        return false;
+    case INPUT_MODE_DIRECT:
+        clear_pending();
+        if (m_commit_pos < m_commitstr.length()) {
+            m_commit_pos = m_commitstr.length();
+            return true;
+        } else {
+            return false;
+        }
+    case INPUT_MODE_PREEDIT:
+        clear_pending();
+        if (m_preedit_pos < m_preeditstr.length()) {
+            m_preedit_pos = m_preeditstr.length();
+        } else if (m_commit_pos < m_commitstr.length()) {
+            m_commit_pos = m_commitstr.length();
+        } else {
+            return false;
+        }
+        return true;
+    default:
+        return false;
+    }
+}
+
 bool
 SKKCore::action_nextpage (void)
 {
@@ -847,6 +953,12 @@ SKKCore::process_remaining_keybinds (const KeyEvent &key)
     if (m_keybind->match_backward_keys(key))
         return action_backward();
 
+    if (m_keybind->match_home_keys(key))
+        return action_home();
+
+    if (m_keybind->match_end_keys(key))
+        return action_end();
+
     return false;
 }
 
@@ -863,6 +975,10 @@ SKKCore::process_ascii (const KeyEvent &key)
         m_keybind->match_convert_keys(key))
         return action_convert();
 
+    if (m_input_mode == INPUT_MODE_PREEDIT &&
+        m_keybind->match_upcase_keys(key))
+        return action_toggle_case();
+
     char code = key.get_ascii_code();
 
     if (!(key.mask & SCIM_KEY_ControlMask || key.mask & SCIM_KEY_Mod1Mask ||
@@ -1030,8 +1146,10 @@ SKKCore::process_key_event (const KeyEvent key)
                 action_select_index(index);
                 return true;
             }
-
-        commit_converting();
+        if (m_cindex != m_candlist.end())
+            commit_converting();
+        else
+            commit_converting(0);
         set_input_mode(INPUT_MODE_DIRECT);
     }
 
@@ -1076,6 +1194,8 @@ SKKCore::process_key_event (const KeyEvent key)
                 /* learning is committed */
                 commit_string(m_learning->m_commitstr);
                 commit_string(m_okuristr);
+                if (m_okurihead != 0)
+                    m_preeditstr += m_okurihead;
                 m_dict->write(m_preeditstr, m_learning->m_commitstr);
                 clear_preedit();
                 clear_candidate();
