@@ -218,6 +218,15 @@ static KeyboardConfigData __config_keyboards_common [] =
         NULL,
         NULL,
     },
+    {
+        SCIM_SKK_CONFIG_UPCASE_KEY,
+        SCIM_SKK_CONFIG_UPCASE_KEY_DEFAULT,
+        N_("Upcase Keys:"),
+        N_("Select Upcase Keys"),
+        N_("The key events to toggle case of characters in preedit string. "),
+        NULL,
+        NULL,
+    },
     {
         NULL,
         "",
@@ -316,6 +325,24 @@ static KeyboardConfigData __config_keyboards_caret [] =
         NULL,
         NULL,
     },
+    {
+        SCIM_SKK_CONFIG_HOME_KEY,
+        SCIM_SKK_CONFIG_HOME_KEY_DEFAULT,
+        N_("Home Keys:"),
+        N_("Select Home Keys"),
+        N_("The key events to move the caret to the start of preedit. "),
+        NULL,
+        NULL,
+    },
+    {
+        SCIM_SKK_CONFIG_BACKWARD_KEY,
+        SCIM_SKK_CONFIG_BACKWARD_KEY_DEFAULT,
+        N_("End Keys:"),
+        N_("Select End Keys"),
+        N_("The key events to move the caret to the end of preedit. "),
+        NULL,
+        NULL,
+    },
     {
         NULL,
         "",
